import numpy as np 
import math
import pandas as pd

class RollingSTD():
    def __init__(self, window_size, days_per_annum, annualize:bool=True):
        self.window_size = window_size
        self.days_per_annum = days_per_annum
        self.annualize = annualize
    
    def compute(self, returns:np.ndarray) -> np.ndarray:
        vol = returns.rolling(self.window_size).std()

        if self.annualize:
            vol = vol * np.sqrt(self.days_per_annum)

        return vol.values

    def predict(self, horizon: int = 1, **kwargs):
        vol = self.compute(**kwargs)
    
        forecast = np.full_like(vol, np.nan, dtype=float)
        forecast[horizon:] = vol[:-horizon]
    
        return forecast    

class EWMA():
    def __init__(self, alpha:float, days_per_annum, span, annualize:bool=True):
        self.alpha = 2 / (span + 1)
        self.days_per_annum = days_per_annum
        self.annualize = annualize
    
        if not 0 < alpha < 1:
            raise ValueError("Alpha must be between 0 and 1")

    def compute(self, returns:np.ndarray):
        r2 = returns**2
        var = np.zeros_like(r2) 

        var[0] = r2[0]
        for i in range(1, len(r2)):
            var[i] = self.alpha * r2[i-1] + (1-self.alpha) * var[i-1]

        vol = np.sqrt(var)
        if self.annualize:
            vol = vol * np.sqrt(self.days_per_annum)
        
        return vol
    
    def predict(self, horizon: int = 1, **kwargs):
        vol = self.compute(**kwargs)

        forecast = np.full_like(vol, np.nan, dtype=float)
        forecast[horizon:] = vol[:-horizon]

        return forecast    


# Downside volatility Upside volatility Semi-variance

import numpy as np

class DownSideVolatility:
    def __init__(self, window: int, annualize: bool = False, days_per_annum: int = 252):
        self.window = window
        self.annualize = annualize
        self.days_per_annum = days_per_annum

    def compute(self, returns:np.ndarray) -> np.ndarray:
        vol = np.full_like(returns, np.nan, dtype = float)

        for i in range(self.window - 1, len(returns)):
            window_slice = returns[i - self.window+1 : i+1]
            negative_returns = window_slice[window_slice < 0]

            if len(negative_returns) > 0:
                vol[i] = np.sqrt(np.mean(negative_returns**2) ) 

        if self.annualize:
            vol = vol * np.sqrt(self.days_per_annum)

        return vol           

    def predict(self, horizon: int = 1, **kwargs):
        vol = self.compute(**kwargs)
    
        forecast = np.full_like(vol, np.nan, dtype=float)
        forecast[horizon:] = vol[:-horizon]
    
        return forecast    


import numpy as np

def future_realized_vol(returns: np.ndarray, horizon: int = 1) -> np.ndarray:
    realized = np.full_like(returns, np.nan, dtype=float)
    n=len(realized)
    for i in range(n - horizon):
        window = returns[i+1 : i+1 + horizon]
        realized[i] = np.sqrt(np.mean(window **2))
        
    return realized


def rmse(predicted: np.ndarray, realized: np.ndarray) -> float:
    mask = ~np.isnan(predicted) & ~np.isnan(realized)
    return np.sqrt(np.mean((predicted[mask] - realized[mask])**2))


def evaluate(model, returns: np.ndarray, 
             compute_kwargs: dict,
             horizon: int = 1) -> float:
    
    forecast = model.predict(horizon=horizon, **compute_kwargs)
    realized = future_realized_vol(returns, horizon=horizon)

    return rmse(forecast, realized)


import numpy as np

class VolOfVol:
    def __init__(self, window: int, annualize: bool = False, days_per_annum: int = 252):
        self.window = window
        self.annualize = annualize
        self.days_per_annum = days_per_annum

    def compute(self, vol_series: np.ndarray) -> np.ndarray:
        vov = np.full_like(vol_series, np.nan, dtype=float)

        for i in range(self.window - 1, len(vol_series)):
            window_slice = vol_series[i - self.window + 1 : i + 1]
            vov[i] = np.std(window_slice, ddof=1)

        if self.annualize:
            vov *= np.sqrt(self.days_per_annum)

        return vov

    def predict(self, horizon: int = 1, **kwargs):
        vol = self.compute(**kwargs)
    
        forecast = np.full_like(vol, np.nan, dtype=float)
        forecast[horizon:] = vol[:-horizon]
    
        return forecast    


__version__ = "0.1.2"
from .realized import RollingSTD, EWMA
from .range_based import Parkinson, GarmanKlass
from .vol_of_vol import VolOfVol
from .asymmetric import DownSideVolatility
from .evaluation import evaluate, rmse


import numpy as np
import math
import pandas as pd


class Parkinson:
    def __init__(self, window: int, annualize: bool = True, days_per_annum: int = 252):
        self.window = window
        self.annualize = annualize
        self.days_per_annum = days_per_annum

    def compute(self, high: np.ndarray, low: np.ndarray) -> np.ndarray:
        log_hl = np.log(high / low)

        daily_var = (1 / (4 * np.log(2))) * (log_hl ** 2)

        rolling_var = pd.Series(daily_var).rolling(self.window).mean()

        vol = np.sqrt(rolling_var)

        if self.annualize:
            vol *= np.sqrt(self.days_per_annum)

        return vol.values

    def predict(self, horizon: int = 1, **kwargs):
        vol = self.compute(**kwargs)
    
        forecast = np.full_like(vol, np.nan, dtype=float)
        forecast[horizon:] = vol[:-horizon]
    
        return forecast    

class GarmanKlass:
    def __init__(self, window: int, annualize: bool = True, days_per_annum: int = 252):
        self.window = window
        self.annualize = annualize
        self.days_per_annum = days_per_annum

    def compute(self, open_: np.ndarray,
                      high: np.ndarray,
                      low: np.ndarray,
                      close: np.ndarray) -> np.ndarray:

        log_hl = np.log(high / low)
        log_co = np.log(close / open_)

        daily_var = (
            0.5 * (log_hl ** 2)
            - (2 * np.log(2) - 1) * (log_co ** 2)
        )

        rolling_var = pd.Series(daily_var).rolling(self.window).mean()

        vol = np.sqrt(rolling_var)

        if self.annualize:
            vol *= np.sqrt(self.days_per_annum)
 
        return vol.values
    
    def predict(self, horizon: int = 1, **kwargs):
        vol = self.compute(**kwargs)
    
        forecast = np.full_like(vol, np.nan, dtype=float)
        forecast[horizon:] = vol[:-horizon]
    
        return forecast    